#include <unistd.h>
#include <errno.h>

#define IOT_BUFSIZE 256
#define IOT_QUIT -42

static struct {
	char quit;              /* set to 1 to get the thread to quit */

	/* output */
	char *data;
	size_t len;

	/* thread */
	SDL_Thread *tid;
	SDL_sem *handled;       /* main to thread: 0: thread's waiting non-blocking after a read; when main finishes picking up it increases it back to 1 to start the next read */
	SDL_sem *enable;        /* main to thread: after a succesful read, wait until enable is 1 to report it */
	unsigned disabled:1;
	SDL_Event ev;
} iot_stdio;

static int mbtk_sdl2_iot_stdio(void *tctx)
{
	sdl2_display_t *sd = tctx;
	char *bufs = malloc(IOT_BUFSIZE*2); /* can't be static because there may be multiple stdio threads for multiple displays */
	int bi;

	iot_stdio.ev.type = mbtk_sdl2_ev_rstdio;
	iot_stdio.ev.user.data1 = sd;

	for(bi = 0;; bi = !bi) {
		int len;
		char *buf = bi ? &bufs[IOT_BUFSIZE] : &bufs[0]; /* need to double buffer because we read in the background while the previous is processed by the user */

		len = read(0, buf, IOT_BUFSIZE-1);
		if (iot_stdio.quit)
			break;

		if (len > 0)
			buf[len] = '\0';
		else
			buf[0] = '\0';

		iot_stdio.data = buf;
		iot_stdio.len = len;

		/* wait while we are in disable */
		SDL_SemWait(iot_stdio.enable);
		if (iot_stdio.quit)
			break;

		SDL_PushEvent(&iot_stdio.ev);

		/* main thread will increase handled to 1 to indicate the event has been processed */
		SDL_SemWait(iot_stdio.handled);
		if (iot_stdio.quit)
			break;
	}

	free(bufs);
	return 0;
}

static void mbtk_sdl2_watch_stdin_read(mbtk_display_t *disp, int enable)
{
	iot_stdio.quit = 0;

	if (enable) {
		if (iot_stdio.tid == NULL) {
			iot_stdio.handled = SDL_CreateSemaphore(0);
			iot_stdio.enable = SDL_CreateSemaphore(0);
			iot_stdio.tid = SDL_CreateThread(mbtk_sdl2_iot_stdio, "read_stdio", disp);
		}

		/* enable reporting */
		SDL_SemTryWait(iot_stdio.enable); /* decrease enable to 0 */
		SDL_SemPost(iot_stdio.enable); /* increase to 1 to enable */
		iot_stdio.disabled = 0;
	}
	else {
		SDL_SemTryWait(iot_stdio.enable); /* decrease enable to 0 so the read thread is blocked */
		iot_stdio.disabled = 1;
	}
}


static void mbtk_sdl2_watch_stdin_uninit(mbtk_display_t *disp)
{
	if (iot_stdio.tid == NULL)
		return;

	iot_stdio.quit = 1;
	SDL_SemPost(iot_stdio.enable);
	SDL_SemPost(iot_stdio.handled);

	/* can't wait for this thread as it will not exit until the next read */
}

/*** fdwatch ***/

typedef struct {
	int fd;
	unsigned for_read:1;
	unsigned for_write:1;
} mbtk_sdl2_fdwacth_req_t;

static int mbtk_sdl2_iot_fdwatch(void *tctx)
{
	sdl2_display_t *sd = tctx;
	mbtk_watch_fd_t wctx = {0};
	mbtk_event_t dst;

	mbtk_watch_fd_set(&wctx, sd->iot_fdwatch.ctrlr, 1, 0);
	sd->iot_fdwatch.ev.type = mbtk_sdl2_ev_fdwatch;
	sd->iot_fdwatch.ev.user.data1 = sd;

	for(;;) {
		long usec = 1000000-1;
		int res;

		res = mbtk_watch_fd_select(&wctx, &dst, &usec, NULL);
		if (res < 0) {
			if (errno == EINTR)
				continue;
			/* something went wrong with an fd, try to blocking-read control to remove stale fds */
			goto read_ctrl;
		}
		if (res == 0)
			continue;

		if (dst.data.fdwatch.fd == sd->iot_fdwatch.ctrlr) {

			mbtk_sdl2_fdwacth_req_t req;
			int len;

			read_ctrl:;

			len = read(sd->iot_fdwatch.ctrlr, &req, sizeof(req));
			if (len != sizeof(req))
				mbtk_error("mbtk_sdl2_iot_fdwatch(): short read or read error on pipe: %d != %d\n", len, (int)sizeof(req));
			else {
				if (req.fd == IOT_QUIT)
					goto quit;
				mbtk_watch_fd_set(&wctx, req.fd, req.for_read, req.for_write);
			}

			continue;
		}

		sd->iot_fdwatch.fd = dst.data.fdwatch.fd;
		sd->iot_fdwatch.readable = dst.data.fdwatch.readable;
		sd->iot_fdwatch.writable = dst.data.fdwatch.writable;

		SDL_PushEvent(&sd->iot_fdwatch.ev);

		/* main thread will increase handled to 1 to indicate the event has been processed */
		SDL_SemWait(sd->iot_fdwatch.handled);
	}

	quit:;
	mbtk_watch_fd_uninit(&wctx);
	return 0;
}

static void mbtk_sdl2_watch_fd(mbtk_display_t *disp, int fd, int for_read, int for_write)
{
	sdl2_display_t *sd = (sdl2_display_t *)disp;
	mbtk_sdl2_fdwacth_req_t req;
	int len;

	if (sd->iot_fdwatch.tid == NULL) {
		mbtk_be_pipe(&sd->iot_fdwatch.ctrlr, &sd->iot_fdwatch.ctrlw);
		sd->iot_fdwatch.handled = SDL_CreateSemaphore(0);
		sd->iot_fdwatch.tid = SDL_CreateThread(mbtk_sdl2_iot_fdwatch, "io_select", sd);
	}

	req.fd = fd;
	req.for_read = for_read;
	req.for_write = for_write;
	len = write(sd->iot_fdwatch.ctrlw, &req, sizeof(req)); /* lazy: assume atomic write */
	if (len != sizeof(req))
		mbtk_error("mbtk_sdl2_watch_fd(): short write or write error on pipe: %d != %d\n", len, (int)sizeof(req));
}



static void mbtk_sdl2_watch_fd_uninit(mbtk_display_t *disp)
{
	sdl2_display_t *sd = (sdl2_display_t *)disp;
	mbtk_sdl2_fdwacth_req_t req;
	int len, status;

	if (sd->iot_fdwatch.tid == NULL)
		return;

	req.fd = IOT_QUIT;
	req.for_write = req.for_read = 0;
	len = write(sd->iot_fdwatch.ctrlw, &req, sizeof(req)); /* lazy: assume atomic write */
	if (len != sizeof(req))
		mbtk_error("mbtk_sdl2_watch_fd(): short write or write error on pipe: %d != %d\n", len, (int)sizeof(req));
	else
		SDL_WaitThread(sd->iot_fdwatch.tid, &status);
}
