static void mbtk_xlib_clipboard_req(xlib_display_t *xd, XEvent *event);

static void mbtk_xlib_clipboard_set(mbtk_display_t *disp, const char *new_content)
{
	xlib_display_t *xd = (xlib_display_t *)disp;
	xlib_win_t *xw = (xlib_win_t *)disp->topwin;

	if (disp->topwin == NULL) {
		mbtk_error("mbtk_xlib_clipboard_set() requires a top window\n");
		return;
	}

	free(xd->clipboard);
	xd->clipboard = mbtk_strdup(new_content);

	XSetSelectionOwner(xd->display, xd->xa.clipboard, xw->window, CurrentTime);
	if (XGetSelectionOwner(xd->display, xd->xa.clipboard) != xw->window)
		mbtk_error("mbtk_xlib_clipboard_set: failed to set selection ownership\n");
}

static void wait_for_xevent(xlib_display_t *xd, int timeout_sec)
{
	fd_set rd;
	struct timeval tv;

	FD_ZERO(&rd);
	FD_SET(xd->fd, &rd);

	tv.tv_sec = timeout_sec;
	tv.tv_usec = 0;
	select(xd->fd+1, &rd, NULL, NULL, &tv);
}

static char *mbtk_xlib_clipboard_get(mbtk_display_t *disp)
{
	xlib_display_t *xd = (xlib_display_t *)disp;
	xlib_win_t *xw = (xlib_win_t *)disp->topwin;
	int n;
	Atom formats[3];
	char *res = NULL;

	formats[0] = xd->xa.utf8_str;
	formats[1] = xd->xa.comp_str;
	formats[2] = XA_STRING;

	for(n = 0; (n < 3) && (res == NULL); n++) {
		unsigned char *data;
		XEvent event;
		Atom actual_type;
		int actual_format, got;
		unsigned long item_cnt, bytes_after;
		double endtime, now;

		/* BUG: this mechanism will fail if there are two backends running in
		   a single thread and we paste in X: we send out a request to get the
		   clipboard but the other client is us in another display that we
		   are not iterating at all. */


		now = endtime = mbtk_be_get_monotime() + 30;
		for(got = 0; now <= endtime; ) {
			XConvertSelection(xd->display, xd->xa.clipboard, formats[n], xd->xa.selection, xw->window, CurrentTime);
			if (XCheckTypedEvent(xd->display, SelectionRequest, &event))
				mbtk_xlib_clipboard_req(xd, &event); /* just in case we own the clipboard - we need to answer the req */
			if (XCheckTypedEvent(xd->display, SelectionNotify, &event)) {
				/* got the event we are looking for */
				got = 1;
				break;
			}

			if (((long)(endtime - now + 1) % 4) == 0)
				mbtk_error("mbtk_xlib_clipboard_get(): clipboard owner is not responding yet, waiting some more...\n");
			wait_for_xevent(xd, 1);
			now = mbtk_be_get_monotime();
		}

		if (!got)
			return NULL;

		if (event.xselection.property == None)
			continue;

		XGetWindowProperty(xd->display, xw->window, event.xselection.property,
			0, LONG_MAX, False, event.xselection.target, &actual_type, &actual_format,
			&item_cnt, &bytes_after, &data);

		if ((event.xselection.target != AnyPropertyType) && (actual_type != event.xselection.target))
			item_cnt = 0;

		if (item_cnt > 0)
			res = mbtk_strdup((char *)data);

		XFree(data);

		XDeleteProperty(xd->display, event.xselection.requestor, event.xselection.property);
	}
	return res;
}

static void mbtk_xlib_clipboard_free(mbtk_display_t *disp, char *str_got)
{
	free(str_got);
}



/*** serve a request from a remote client ***/

static Atom set_sel_prop(xlib_display_t *xd, const XSelectionRequestEvent *request)
{
	int n;
	long cliplen;
	Atom formats[3];

	formats[0] = xd->xa.utf8_str;
	formats[1] = xd->xa.comp_str;
	formats[2] = XA_STRING;

	if (request->property == None) /* request from legacy client (ICCCM section 2.2) */
		return None;

	if (request->target == xd->xa.savetargets) { /* asks whether we support SAVE_TARGETS */
		XChangeProperty(xd->display, request->requestor, request->property, xd->xa.null, 32, PropModeReplace, NULL, 0);
		return request->property;
	}

	if (request->target == xd->xa.targets) {
		Atom stargets[4]; /* targets we support */

		stargets[0] = xd->xa.targets;
		stargets[1] = formats[0];
		stargets[2] = formats[1];
		stargets[3] = formats[2];

		XChangeProperty(xd->display, request->requestor, request->property,
			XA_ATOM, 32, PropModeReplace, (unsigned char *)stargets, 4);

		return request->property;
	}

	if (xd->clipboard == NULL) /* shouldn't happen, but just in case... */
		return None;

	cliplen = strlen(xd->clipboard);

	if (request->target == xd->xa.mult) {
		Atom *targets, actual_type;
		int actual_format, m;
		unsigned long item_cnt, bytes_after;

		XGetWindowProperty(xd->display, request->requestor, request->property,
			0, LONG_MAX, False, xd->xa.pair, &actual_type, &actual_format, &item_cnt,
			&bytes_after, (unsigned char **)&targets);

		if (actual_type != xd->xa.pair)
			item_cnt = 0;

		for(n = 0; n < item_cnt; n += 2) {
			targets[n+1] = None;
			for(m = 0; m < 3; m++) {
				if (targets[n] == formats[m]) {
					XChangeProperty(xd->display, request->requestor, targets[n+1], targets[n], 8, PropModeReplace, (unsigned char *)xd->clipboard, cliplen);
					break;
				}
			}
		}

		XChangeProperty(xd->display, request->requestor, request->property, xd->xa.pair, 32, PropModeReplace, (unsigned char *)targets, item_cnt);
		XFree(targets);
		return request->property;
	}


	/* convert to requested type */
	for(n = 0; n < 3; n++) {
		if (request->target == formats[n]) {
			XChangeProperty(xd->display, request->requestor, request->property, request->target, 8, PropModeReplace, (unsigned char *)xd->clipboard, cliplen);
			return request->property;
		}
	}

	return None; /* req not supported */
}

static void mbtk_xlib_clipboard_req(xlib_display_t *xd, XEvent *event)
{
	const XSelectionRequestEvent *request = &event->xselectionrequest;
	XEvent reply = {0};

	reply.xselection.property = set_sel_prop(xd, request);
	reply.xselection.type = SelectionNotify;
	reply.xselection.display = request->display;
	reply.xselection.requestor = request->requestor;
	reply.xselection.selection = request->selection;
	reply.xselection.target = request->target;
	reply.xselection.time = request->time;

	XSendEvent(xd->display, request->requestor, False, 0, &reply);
}
