#include <X11/Xutil.h>

#define MAYBE_SKIP_RENDER(skip_stmt) \
do { \
	xlib_win_t *sxw = (xlib_win_t *)win; \
	if (sxw->skip_render) { skip_stmt; } \
} while(0)

static int mbtk_xlib_draw_color(mbtk_window_t *win, int r, int g, int b, int a)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;
	XColor clr;
	int succ;

	MAYBE_SKIP_RENDER(return 0);

	clr.red = (unsigned int)r << 8;
	clr.green = (unsigned int)g << 8;
	clr.blue = (unsigned int)b << 8;
	clr.flags = DoRed | DoGreen | DoBlue;
	succ = XAllocColor(xd->display, xd->cmap, &clr);

	if (!succ)
		return -1;

	XSetForeground(xd->display, xw->gc, clr.pixel);
	return 0;
}

static void mbtk_xlib_draw_fillrect(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	XFillRectangle(xd->display, xw->window, xw->gc, x1, y1, x2 - x1 + 1, y2 - y1 + 1);
}

static void mbtk_xlib_draw_line(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	XDrawLine(xd->display, xw->window, xw->gc, x1, y1, x2, y2);
}

static void mbtk_xlib_draw_hline(mbtk_window_t *win, int x1, int y1, int x2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	if (x1 < x2)
		XDrawLine(xd->display, xw->window, xw->gc, x1, y1, x2, y1);
	else
		XDrawLine(xd->display, xw->window, xw->gc, x2, y1, x1, y1);
}

static void mbtk_xlib_draw_vline(mbtk_window_t *win, int x1, int y1, int y2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	if (y1 < y2)
		XDrawLine(xd->display, xw->window, xw->gc, x1, y1, x1, y2);
	else
		XDrawLine(xd->display, xw->window, xw->gc, x1, y2, x1, y1);
}

int mbtk_xlib_draw_set_font(mbtk_window_t *win, mbtk_font_id_t fontid)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;
	XFontStruct *fnt;

	if ((fontid.l < 0) || (fontid.l >= xd->fonts.used))
		return -1;

	MAYBE_SKIP_RENDER(return 0);

	fnt = xd->fonts.array[fontid.l];
	if (fnt == NULL)
		return -1;

	xw->fnt = fnt;
	XSetFont(xd->display, xw->gc, fnt->fid);

	return 0;
}

static void mbtk_xlib_draw_query_text(mbtk_window_t *win, int *wo, int *asco, int *desco, char *s, int slen)
{
	xlib_win_t *xw = (xlib_win_t *)win;

	if (xw->fnt != NULL) {
		XCharStruct xcs;
		int dir, asc, desc;

		XTextExtents(xw->fnt, s, slen, &dir, &asc, &desc, &xcs);
/*		*wo = XTextWidth(xw->fnt, s, slen);*/
		*wo = xcs.width;
		*asco = asc;
		*desco = desc;
	}
	else {
		*wo = slen << 3;
		*asco = *desco = 5;
	}
}

static void mbtk_xlib_draw_text(mbtk_window_t *win, int x, int y, char *s, int slen, void **cache)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	if (xw->fnt != NULL)
		XDrawString(xd->display, xw->window, xw->gc, x, y, s, slen);
}

static int mbtk_xlib_draw_import_pixmap(mbtk_display_t *display, mbtk_pixmap_t *dst, unsigned char *pixels, int width, int height, int has_alpha)
{
	xlib_display_t *xd = (xlib_display_t *)display;
	Visual *vis = DefaultVisual(xd->display, 0);

	dst->bedata.p = dst->bedata.raw = NULL; /* make sure to keep these invalid if conversion fails */
	return create_ximage(dst, xd->display, DefaultRootWindow(xd->display), vis, xd->depth, pixels, width, height, has_alpha);
}

static void mbtk_xlib_draw_free_pixmap(mbtk_display_t *display, mbtk_pixmap_t *px)
{
	xlib_display_t *xd = (xlib_display_t *)display;
	XImage *i = px->bedata.p;
	Pixmap m = px->bedata.ul;

	if (m != 0)
		XFreePixmap(xd->display, m);

	XDestroyImage(i);

/*	free(px->bedata.raw); - XDestroyImage() does this */
	px->bedata.p = px->bedata.raw = NULL;
}

static mbtk_px_id_t mbtk_xlib_draw_bind_pixmap(mbtk_window_t *win, mbtk_pixmap_t *px)
{
	mbtk_px_id_t pxid;
	pxid.p = px;
	pxid.l = 1; /* make it look valid */
	return pxid;
}

static void mbtk_xlib_draw_pixmap(mbtk_window_t *win, int x, int y, mbtk_px_id_t pxid)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;
	mbtk_pixmap_t *px = pxid.p;
	XImage *img = px->bedata.p;
	Pixmap mask = px->bedata.ul;

	MAYBE_SKIP_RENDER(return);

	if (mask != 0) {
		XSetClipMask(xd->display, xw->gc, mask);
		XSetClipOrigin(xd->display, xw->gc, x, y);
		XPutImage(xd->display, xw->window, xw->gc, img, 0, 0, x, y, px->w, px->h);
		XSetClipMask(xd->display, xw->gc, None);
	}
	else
		XPutImage(xd->display, xw->window, xw->gc, img, 0, 0, x, y, px->w, px->h);
}


static void mbtk_xlib_draw_flush(mbtk_window_t *win)
{
	xlib_display_t *xd = (xlib_display_t *)win->display;
	xlib_win_t *xw = (xlib_win_t *)win;

	if (win->in_widget != NULL) {
		if (win->native_draw_post != NULL)
			win->native_draw_post(win, &xw->window);
		return;
	}

	XFlush(xd->display);
}


static void mbtk_xlib_draw_start(mbtk_window_t *win)
{
	if (win->in_widget != NULL) {
		xlib_win_t *sxw = (xlib_win_t *)win;
		sxw->skip_render = 0;
		if (win->native_draw_pre != NULL)
			sxw->skip_render = win->native_draw_pre(win, &sxw->window);
	}
}



