#define GENPRIQUE_PRI_TYPE double
#define GENPRIQUE_API MBTK_INLINE

#include "libmbtk.h"
#include <libmbtk/src_3rd/genprique/fibheap.h>
#include <libmbtk/src_3rd/genprique/fibheap.c>

#include "timer.h"

#define MBTK_TIMER_REPEAT_DELETE (-3)


typedef struct mbtk_timer_s {
	mbtk_timer_id_t id;
	long rep;
	double period;
	mbtk_timer_res_t (*cb)(long repleft, void *user_data);
	void *user_data;
	fbhn_t fbhn;        /* ->pri is the next time to run */
} mbtk_timer_t;

/* Allocate a new timer struct; all fields are cleared to zero except for id,
   which is loaded with the timer id */
static mbtk_timer_t *mbtk_timer_alloc(mbtk_display_t *disp)
{
	long n;
	mbtk_timer_t *timer;

	if (disp->timerq == NULL) {
		disp->timerq = malloc(sizeof(fbh_t));
		if (fbh_init(disp->timerq, offsetof(mbtk_timer_t, fbhn)) != 0) {
			free(disp->timerq);
			return NULL;
		}
	}

	timer = calloc(sizeof(mbtk_timer_t), 1);

	/* find a free id */
	for(n = 0; n < disp->timers.used; n++) {
		if (disp->timers.array[n] == NULL) {
			timer->id = n;
			return timer;
		}
	}

	/* not found, allocate new id */
	timer->id = disp->timers.used;
	vtp0_append(&disp->timers, timer);
	return timer;
}

static void mbtk_timer_free(mbtk_display_t *disp, mbtk_timer_id_t id, int optimize)
{
	free(disp->timers.array[id]);
	disp->timers.array[id] = NULL;

	if (optimize && (id == disp->timers.used-1)) {
		long n;
		/* remove trailing NULLs */
		for(n = disp->timers.used-1; n >= 0; n--)
			if (disp->timers.array[n] == NULL)
				disp->timers.used--;
	}
}

void mbtk_timer_uninit_display(mbtk_display_t *disp)
{
	long n;

	if (disp->timerq != NULL)
		free(disp->timerq);

	for(n = 0; n < disp->timers.used; n++)
		mbtk_timer_free(disp, n, 0);
	vtp0_uninit(&disp->timers);

}

mbtk_timer_id_t mbtk_timer_new(mbtk_display_t *disp, long rep, double period, mbtk_timer_res_t (*cb)(long repleft, void *user_data), void *user_data)
{
	mbtk_timer_t *t = mbtk_timer_alloc(disp);
	double now = mbtk_get_time_el(disp);

	t->rep = rep;
	t->period = period;
	t->cb = cb;
	t->user_data = user_data;

	fbh_insert(disp->timerq, t, now + period);

	return t->id;
}

double mbtk_timer_next(mbtk_display_t *disp, double now)
{
	fbh_t *fbh = disp->timerq;
	mbtk_timer_t *t;
	double delta;

	if (fbh == NULL)
		return -1;

	t = fbh_min(fbh);
	if (t == NULL)
		return -1;

	delta = t->fbhn.pri - now;
	if (delta < 0)
		return 0;
	return delta;
}

void mbtk_timer_run(mbtk_display_t *disp, double now)
{
	fbh_t *fbh = disp->timerq;

	if (fbh == NULL)
		return;

	for(;;) {
		mbtk_timer_t *t = fbh_min(fbh);
		mbtk_timer_res_t r;

		if (t == NULL)
			return; /* no active timer */

		if (t->fbhn.pri > now)
			return; /* hottest timer expires in the future, nothing else to do here */

		fbh_pop_min(fbh);

		if (t->rep != MBTK_TIMER_REPEAT_DELETE) {
			if (t->rep > 0)
				t->rep--;

			r = t->cb(t->rep, t->user_data);
			if (t->rep == 0)
				r = MBTK_TIMER_STOP;
		}
		else
			r = MBTK_TIMER_STOP;

		if (r == MBTK_TIMER_CONTINUE)
			fbh_insert(disp->timerq, t, now + t->period);
		else
			mbtk_timer_free(disp, t->id, 1);
	}
}


#define CHKID(disp, id, error_stmt) \
do { \
	if ((id < 0) || (id >= disp->timers.used) || (disp->timers.array[tid] == NULL)) { error_stmt; } \
} while(0)

int mbtk_timer_stop(mbtk_display_t *disp, mbtk_timer_id_t tid)
{
	mbtk_timer_t *t;

	CHKID(disp, tid, return -1);

	/* can't remove random item from the prique; rather mark the timer to be
	   deleted and remove it when it expires */
	t = disp->timers.array[tid];
	t->rep = MBTK_TIMER_REPEAT_DELETE;
	return 0;
}


long mbtk_timer_get_repleft(mbtk_display_t *disp, mbtk_timer_id_t tid)
{
	mbtk_timer_t *t;

	CHKID(disp, tid, return -1);

	t = disp->timers.array[tid];
	if (t->rep == MBTK_TIMER_REPEAT_DELETE)
		return -1;

	return t->rep;
}

