#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/tabbed.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/label.h>
#include <libmbtk/widget/Common.h>

/*
Design: this widget is a VBOX of two widgets. The first child is a HBOX,
the tab list: each widget is a tab selector (typically a label). The second
child is a HBOX with one widget per page.

The selector is handled by hand_radio to make sure only one widget is active.
All pages are hidden, except the one whose selector is active.

The above is for the normal, 'top' alignemnt, where tabs are in a horizontal
line on the top. When alignment is 'left', the main box is a HBOX and the
selector box is a VBOX. (Orientation of the pages box doesn't matter as only
one widget of it is visible at a time.)
*/

static void mbtk_tabbed_imp_destroy(mbtk_widget_t *w)
{
	mbtk_tabbed_t *tabbed = (mbtk_tabbed_t *)w;
	mbtk_handler_radio_uninit(&tabbed->radio);
}


static const mbtk_widget_imp_t mbtk_tabbed_imp = {
	mbtk_tabbed_imp_destroy,

	mbtk_comm_frame_fallback_minsize,
	mbtk_comm_frame_fallback_draw,
	mbtk_comm_frame_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL  /* pre_unassign */
};

mbtk_tabbed_t *mbtk_tabbed_new(mbtk_tabbed_t *caller_alloced, mbtk_tabbed_layout_t layout)
{
	mbtk_tabbed_t *tabbed;
	mbtk_kw_t kw_tabbed;

	mbtk_kw_cache(kw_tabbed, "tabbed");

	mbtk_create_wt(tabbed, kw_tabbed, &mbtk_tabbed_imp, caller_alloced);
	tabbed->w.hvbox.type = (layout == MBTK_TABBED_TOP) ? HVBOX_VBOX : HVBOX_HBOX;

	mbtk_box_new(&tabbed->selectors, (layout == MBTK_TABBED_TOP));
	mbtk_box_add_widget((mbtk_box_t *)tabbed, &tabbed->selectors.w, 0);
	mbtk_box_set_align(&tabbed->selectors, HVBOX_LEFT_TOP);
	mbtk_box_set_spacing(&tabbed->selectors, 4);

	mbtk_box_new(&tabbed->pages, 0);
	mbtk_box_add_widget((mbtk_box_t *)tabbed, &tabbed->pages.w, 0);

	mbtk_handler_radio_init(NULL, &tabbed->radio); /* win == NULL is okay here: we are going to free the handler on widget destroy */

	/*
		tabbed->w.hvbox.widget = "tabbed";
		tabbed->selectors.w.hvbox.widget = "tabbed.selectors";
		tabbed->pages.w.hvbox.widget = "tabbed.pages";
	*/

	return tabbed;
}

static mbtk_event_handled_t mbtk_tabbed_radio_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	mbtk_tabbed_t *tabbed = user_data;
	hvbox_t *selectors = tabbed->w.hvbox.ch_first, *pages = selectors->next;
	hvbox_t *sl, *pg;

	/* find which selector is activated, show correspoding page, hide all other pages */
	for(sl = selectors->ch_first, pg = pages->ch_first; sl != NULL; sl = sl->next, pg = pg->next) {
		mbtk_widget_t *wsl = (mbtk_widget_t *)sl, *wpg = (mbtk_widget_t *)pg;
		if (wsl == w)
			mbtk_widget_show(wpg, 0);
		else
			mbtk_widget_hide(wpg, 0);
	}

	return MBTK_EVENT_HANDLED;
}


void mbtk_tabbed_append_page(mbtk_tabbed_t *tabbed, mbtk_widget_t *selector, mbtk_widget_t *page)
{
	mbtk_kw_t kw_toggled;
	hvbox_t *selectors = tabbed->w.hvbox.ch_first, *pages = selectors->next;

	mbtk_box_add_widget((mbtk_box_t *)selectors, selector, 0);
	mbtk_box_add_widget((mbtk_box_t *)pages, page, 0);

	page->hvbox.hidden = 1;
	selector->flags.activated = 0;

	mbtk_widget_embedded((mbtk_widget_t *)selector->hvbox.parent, selector);
	mbtk_widget_embedded((mbtk_widget_t *)page->hvbox.parent, page);

	mbtk_handler_radio_add_widget(&tabbed->radio, selector);
	mbtk_kw_cache(kw_toggled, "toggled");
	mbtk_callback_set_noarg(selector, kw_toggled, mbtk_tabbed_radio_cb, tabbed);
}

void mbtk_tabbed_append_page_label(mbtk_tabbed_t *tabbed, const char *selector, mbtk_widget_t *page)
{
	mbtk_label_t *lab = mbtk_label_new(NULL, selector);
	lab->draw_bg = 1;
	mbtk_tabbed_append_page(tabbed, &lab->w, page);
}


mbtk_widget_t *mbtk_tabbed_subwidget(mbtk_tabbed_t *tabbed, mbtk_kw_t subname)
{
	hvbox_t *selectors = tabbed->w.hvbox.ch_first;
	mbtk_kw_t kw_selectors, kw_pages;

	mbtk_kw_cache(kw_selectors, "selectors");
	mbtk_kw_cache(kw_pages, "pages");

	if (subname == kw_selectors) return (mbtk_widget_t *)selectors;
	if (subname == kw_pages) return (mbtk_widget_t *)selectors->next;

	return NULL;
}
